@@ -3,6 +3,7 @@ from math import exp, pi
 import os
 import numpy as np
 import pytest
+import itertools as it
 
 from pycvodes import (
     integrate_adaptive, integrate_predefined, requires_jac, get_include
@@ -60,11 +61,62 @@ def _get_f_j(k):
             dfdx_out[2] = 0
     return f, j
 
+
+def _get_f_j_sparse(k):
+    k0, k1, k2 = k
+
+    def f(t, y, fout):
+        fout[0] = -k0*y[0]
+        fout[1] = k0*y[0] - k1*y[1]
+        fout[2] = k1*y[1] - k2*y[2]
+
+    _colptrs = [0, 2, 4, 5]
+    _rowvals = [0, 1, 1, 2, 2]
+
+    def j(t, y, data, colptrs, rowvals):
+        data[:] = -k0, k0, -k1, k1, -k2
+        colptrs[:] = _colptrs
+        rowvals[:] = _rowvals
+
+    return f, j, len(_rowvals)
+
+
+def _gravity_f_j_jtimes(g):
+    """ RHS, jacobian, and jacobian-vector product for
+        particle falling under influence of gravity"""
+
+    def f(t, y, fout):
+        x, v = y
+        fout[0] = v
+        fout[1] = -g
+
+    def jac(t, y, jmat_out, dfdx_out=None, fy=None):
+        jmat_out[0, 0] = 0
+        jmat_out[0, 1] = 1
+        jmat_out[1, 0] = 0
+        jmat_out[0, 1] = 0
+
+    def jtimes(v, Jv, t, y, fy, user_data=None, tmp=None):
+        Jv[0] = v[1]
+        Jv[1] = 0
+
+    return f, jac, jtimes
+
+
+def gravity_analytic(g, y0, tout):
+    """ Analytic solution to the above at times tout"""
+    x0, v0 = y0
+    t0 = tout[0]
+    return np.column_stack((x0 - 0.5 * g * (tout - t0)**2, v0 - g * (tout - t0)))
+
+
 methods = [('adams', 1.8, False),
            ('adams', 1.8, True),
            ('bdf', 10, False),
            ('bdf', 10, True)]
 
+iterative_linear_solvers = ["gmres", "bicgstab", "tfqmr"]
+
 
 def bandify(cb, mlower, mupper):
     def j(t, y, jmat_out, dfdx_out=None, fy=None):
@@ -537,3 +589,83 @@ def test_adaptive_ew_ele():
     assert info['success']
     abs_ew_ele = np.abs(np.prod(info['ew_ele'], axis=1))
     assert np.all(abs_ew_ele < 1)
+
+
+jtimes_params = list(it.product(iterative_linear_solvers, [True, False]))
+
+
+@pytest.mark.parametrize("linear_solver,with_jac", jtimes_params)
+def test_jtimes_adaptive(linear_solver, with_jac):
+    g = 9.81
+    y0 = [1000.0, 0.0]
+    atol, rtol = 1e-8, 1e-8
+    f, jac, jtimes = _gravity_f_j_jtimes(g)
+    if not with_jac:
+        jac = None
+    kwargs = dict(atol=atol, rtol=rtol,
+                  method='bdf', linear_solver=linear_solver,
+                  jtimes=jtimes)
+    tout, yout, info = integrate_adaptive(f, jac, y0, 0, 10, **kwargs)
+    yref = gravity_analytic(g, y0, tout)
+    assert np.allclose(yout, yref, rtol=10*rtol, atol=10*atol)
+    assert info['success']
+    assert info['njvev'] > 0
+    if not with_jac:
+        assert info['njev'] == 0
+
+
+@pytest.mark.parametrize("linear_solver,with_jac", jtimes_params)
+def test_jtimes_predefined(linear_solver, with_jac):
+    g = 9.81
+    y0 = [1000.0, 0.0]
+    atol, rtol = 1e-8, 1e-8
+    f, jac, jtimes = _gravity_f_j_jtimes(g)
+    if not with_jac:
+        jac = None
+    kwargs = dict(atol=atol, rtol=rtol,
+                  method='bdf', linear_solver=linear_solver,
+                  jtimes=jtimes)
+    tout = np.linspace(0, 10, 100)
+    yout, info = integrate_predefined(f, jac, y0, tout, **kwargs)
+    yref = gravity_analytic(g, y0, tout)
+    assert np.allclose(yout, yref, rtol=10*rtol, atol=10*atol)
+    assert info['success']
+    assert info['njvev'] > 0
+    if not with_jac:
+        assert info['njev'] == 0
+
+
+def test_sparse_jac_adaptive():
+    from .._config import env
+    if env['LAPACK'] in ('', '0'):
+        pytest.skip("sparse jacobian tests require BLAS/LAPACK for KLU solver")
+    k = 2.0, 3.0, 4.0
+    y0 = [0.7, 0.3, 0.5]
+    f, j, nnz = _get_f_j_sparse(k)
+    atol, rtol = 1e-8, 1e-8
+    kwargs = dict(atol=atol, rtol=rtol, method='bdf',
+                  linear_solver='klu', nnz=nnz)
+    xout, yout, info = integrate_adaptive(f, j, y0, 0, 10, **kwargs)
+    yref = decay_get_Cref(k, y0, xout - xout[0])
+    assert info['success']
+    assert info['njev'] > 0
+    assert np.allclose(yout, yref, rtol=10*rtol, atol=10*atol)
+
+
+def test_sparse_jac_predefined():
+    from .._config import env
+    if env['LAPACK'] in ('', '0'):
+        pytest.skip("sparse jacobian tests require BLAS/LAPACK for KLU solver")
+    else:
+        print("env['LAPACK'] = %s" % env['LAPACK'])  # debug
+    k = 2.0, 3.0, 4.0
+    y0 = [0.7, 0.3, 0.5]
+    xout = np.linspace(0, 3, 31)
+    f, j, nnz = _get_f_j_sparse(k)
+    atol, rtol = 1e-8, 1e-8
+    kwargs = dict(method='bdf', linear_solver='klu', nnz=nnz)
+    yout, info = integrate_predefined(f, j, y0, xout, 1.0e-8, 1.0e-8, **kwargs)
+    yref = decay_get_Cref(k, y0, xout - xout[0])
+    assert info['success']
+    assert info['njev'] > 0
+    assert np.allclose(yout, yref, rtol=10*rtol, atol=10*atol)
